/*--------------------------------------------*/
/*        Written by Siddharth Agarwal        */
/*--------------------------------------------*/
/*----------------------------------------------------*/
/* This function calculates the value of Pi square*/
/*----------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

double vol(double x, double y);
double calpi(double inix, double finx, double iniy, double finy, double h);
main(int argc, char* argv[])
{
 int p;
 int dim_sizes[2];
 int wrap_around[2];
 int reorder = 1;
 int coordinates[2];
 int rank ,q,i,j;
 double fin[10]; /* this stores the values of the pi squares calculated by the individual processors.*/
 double x, y;
 double inix; /* initilial coordinate of x*/
 double finx; /* final coordinate of x*/
 double iniy; /* initial coordinate of y*/
 double finy; /* initial coordinate of y*/
 double h;
 double pisqr; /* the value of pi squre that each processor calculates*/
 double final = 0.0; /* The final value of pi square after all the values are added together */
 

 MPI_Comm grid_comm;
 /* This function starts up MPI, it must be called before any other MPI functio is called */
 MPI_Init(&argc, &argv);

 /* Returns the no of processors in the underlying comm */
    MPI_Comm_size(MPI_COMM_WORLD, &p);

    q = sqrt(p);
    dim_sizes[0] = dim_sizes[1] = q;
    wrap_around[0] = wrap_around[1] = 1;
 
 /* creates a new communicator with cartesian coordinate structure on 
    the processes*/  
    MPI_Cart_create(MPI_COMM_WORLD, 2, dim_sizes, wrap_around, reorder, &grid_comm);

 /* Returns the rank of the calling process in the group underlying the
    communicator comm */
    MPI_Comm_rank(grid_comm, &rank);

 /* Returns the coordinates of the process with rank rank in the
    cartesian communicator comm */ 
    MPI_Cart_coords(grid_comm, rank, 2, coordinates);
    

    h = 1/(double)p;
    inix = (double)coordinates[0]/(double)q; /* Calculates the initial coordinate of x*/
    iniy = (double)coordinates[1]/(double)q; /* Calculates the initial coordinate of y*/
    finx = (double)coordinates[0]/(double)q + (1/(double)q);  /* Calculates the final coordinate of x*/
    finy = (double)coordinates[1]/(double)q + (1/(double)q);  /* Calculates the final coordinate of y*/
    for (x=inix; x<finx; x+=h)
    { 
     for (y=iniy; y<finy; y+=h)
     { 
      pisqr += (h*h) * (vol(x,y) + vol(x,y+h) + vol(x+h,y) + vol(x+h,y+h))/4;
     }
    }
   pisqr = 16 * pisqr;
   
   MPI_Gather(&pisqr, 1, MPI_DOUBLE, &fin, 1, MPI_DOUBLE, 0, grid_comm);

  /* If the rank  of the process is 0 then it adds all the values and gives the final pi square value*/
   if(rank == 0)
   { 
    for(i = 0; i < p; i++)
    final += fin[i];
    printf("The  Final value of pi squared is %.16f\n", final); 
   }

   /* this function shuts down MPI. No MPI function may be called after this function */
   MPI_Finalize();

 }

 double vol(double x, double y)
 { 
  return (double)1/((1.0 +x*x) *(1.0 + y*y));
 }

/* double calpi(double inix,double finx,double iniy,double finy,double h)
 { 
  double x, y;
  double temp;
  for (x=inix; x<finx; x+=h)
    { 
     for (y=iniy; y<finy; y+=h)
     { 
      temp += (h*h) * (vol(x,y) + vol(x,y+h) + vol(x+h,y) + vol(x+h,y+h))/4;
     }
    }
   temp = 16 * temp; 
  return temp;
}*/
